/**
 * ViewPort.js - Code to create a viewport window in a web browser.
 *
 * Requires prototypejs 1.6+ and scriptaculous 1.8+
 * Requires Array.numericSort() and Hash.flip() functions (defined in
 * DimpCore.js).
 * Requires DimpSlider.js.
 *
 * $Horde: dimp/js/src/ViewPort.js,v 1.194.2.6 2008/01/18 06:57:52 slusarz Exp $
 *
 * Copyright 2005-2008 The Horde Project (http://www.horde.org/)
 *
 * See the enclosed file COPYING for license information (GPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/gpl.html.
 */

/**
 * ViewPort
 */
var ViewPort = Class.create({

    // Required options: content_container, lines, fetch_action, template,
    //                   cachecheck_action, ajaxRequest, buffer_pages,
    //                   limit_factor, content_class, row_class, maxHeight,
    //                   msgid_prefix
    initialize: function(opts)
    {
        opts.content = $(opts.content_container);
        opts.empty = opts.empty_container ? $(opts.empty_container) : null;
        opts.error = opts.error_container ? $(opts.error_container) : null;
        this.opts = opts;

        this.scroller = new ViewPort_Scroller(this);
        this.template = new Template(opts.template);
        this.current_requests = $H();
        this.views = $H();

        // Initialize all other variables
        this.busyrequest = this.firstcontent = this.force = this.id = this.isbusy = this.line_height = this.page_size = this.show_split_pane = this.splitbar = this.splitbar_loc = this.viewport_init = null;
        this.last_page_size = this.last_start_pos = -1;
    },

    loadView: function(id, params, uid, background)
    {
        var init, opts = {};
        this.clearWait();

        if (Object.isUndefined(this.page_size)) {
            this.page_size = (this.show_split_pane) ? this.getDefaultPageSize() : this.getMaxPageSize();
        }

        if (this.buffer) {
            this.views.set(this.id, { buffer: this.buffer, offset: this.currentOffset() });
            if (this.viewport_init != 2) {
                this.viewport_init = 1;
                this.renderViewport();
            }
            if (this.opts.onViewChange) {
                this.opts.onViewChange();
            }
        } else {
            init = true;
        }

        this.additional_params = $H(params);
        this.id = id;

        if (background) {
            opts.noupdate = true;
        }

        var curr = this.views.get(id);
        if (curr) {
            this.buffer = curr.buffer;
            this.updateContent(curr.offset, opts);
            this.checkCache();
            if (this.opts.onComplete) {
                this.opts.onComplete();
            }
            return true;
        }

        this.buffer = new ViewPort_Buffer(this, this.opts.buffer_pages, this.opts.limit_factor);

        if (init && !background) {
            this.viewport_init = 1;
            this.renderViewport();
        }

        this.fetchBuffer((uid) ? 'uid' : 'offset', (uid) ? uid : 0, opts);
        this.clearWait();
        return false;
    },

    deleteView: function(id)
    {
        this.views.unset(id);
    },

    getRowById: function(id, view)
    {
        var buffer = this.getBuffer(view);
        return (buffer) ? buffer.getRowById(id) : null;
    },

    inBuffer: function(id, view)
    {
        var buffer = this.getBuffer(view);
        return (buffer) ? buffer.inBuffer(id) : false;
    },

    getBuffer: function(view) {
        return (view && view != this.id) ? (this.views.get(view) ? this.views.get(view).buffer : null) : this.buffer;
    },

    positionById: function(id)
    {
        var row = this.getRowById(id);
        if (!row) {
            return false;
        }

        switch (this.visible(id)) {
        case -1:
            this.scroller.moveScroll(row.rownum - 1);
            break;

        // case 0:
        // noop

        case 1:
            this.scroller.moveScroll(row.rownum - this.getPageSize());
        }

        return true;
    },

    visible: function(id)
    {
        var row = this.getRowById(id);
        if (!row) {
            return false;
        }

        if (row.rownum < this.last_start_pos + 1) {
            return -1;
        } else {
            return (row.rownum > (this.last_start_pos + this.getCurrPageSize())) ? 1 : 0;
        }
    },

    reload: function(params)
    {
        if (this.filter && this.filter.isFiltering()) {
            this.filter.filter(null, params);
        } else {
            this.fetchBuffer('offset', this.currentOffset(), null, $H(params).merge({ update: 1 }).toObject());
        }
    },

    refresh: function()
    {
        if (!this.requestContentRefresh(this.currentOffset()) &&
            this.opts.onComplete) {
            // Even if there has not been a change to the current view, other
            // viewport parameters may have changed so we need to call
            // onComplete() to give other elements a chance to refresh.
            this.opts.onComplete();
        }
    },

    remove: function(uids, cacheid)
    {
        this.setMetaData({ cacheid: cacheid });

        var visible = uids.intersect(this.buffer.getRows(this.last_start_pos + 1, this.last_page_size).pluck('uid'));

        // If we have visible elements to remove, only call refresh
        // after the last effect has finished.
        this.isbusy = false;
        var vsize = visible.size();
        if (vsize) {
            var vrows = [];
            for (var i = 0; i < vsize; ++i) {
                this.isbusy = true;
                // Set 'to' to a value slightly above 0 to prevent Effect.Fade
                // from auto hiding.  Hiding is unnecessary, since we will be
                // removing from the document shortly.
                var args = { duration: 0.4, to: 0.01 };
                if ((i + 1) == vsize) {
                    args.afterFinish = this.removeids.bind(this, uids, true);
                }
                vrows.push(this.getRowDiv(visible[i]));
                Effect.Fade(vrows.last(), args);
            }
            if (this.opts.onClearRows) {
                this.opts.onClearRows(vrows);
            }
            this.last_page_size -= vsize;
        } else {
            this.removeids(uids);
        }
    },

    removeids: function(uids, visible)
    {
        var rows = (this.opts.onRemoveRows) ? uids.collect(function(id) { return this.getRowById(id); }.bind(this)).compact() : [];

        this.buffer.removeByUid(uids);
        this.setMetaData({ total_rows: this.getMetaData('total_rows') - uids.size() });

        // We need to update the scrollbar size here or else the offset will
        // be computed wrong in refresh().
        if (visible) {
            this.scroller.updateSize();
        }
        this.refresh();

        if (this.opts.onRemoveRows) {
            this.opts.onRemoveRows(rows);
        }

        this.isbusy = false;
    },

    addFilter: function(action, callback)
    {
        this.filter = new ViewPort_Filter(this, action, callback);
    },

    runFilter: function(val, params)
    {
        if (this.filter) {
            this.filter.filter(Object.isUndefined(val) ? null : val, params);
        }
    },

    // reset = (boolean) If true, don't update the viewport
    stopFilter: function(reset)
    {
        if (this.filter) {
            this.filter.clear(reset);
        }
    },

    onResize: function(noupdate)
    {
        if (!this.firstcontent || !this.opts.content.visible()) {
            return;
        }

        if (this.opts.onBeforeResize) {
            this.opts.onBeforeResize();
        }

        this.renderViewport(noupdate);

        if (this.opts.onAfterResize) {
            this.opts.onAfterResize();
        }
    },

    requestContentRefresh: function(offset)
    {
        if (offset == this.last_start_pos &&
            this.getPageSize() == this.last_page_size) {
            return false;
        }
        if (this.buffer.isLoaded(offset)) {
            this.updateContent(offset);
            var limit = this.buffer.isNearingLimit(offset);
            if (limit) {
                this.fetchBuffer('offset', offset, { noupdate: true }, { nearing: limit, page_size: this.getPageSize() });
            }
            return true;
        } else {
            this.fetchBuffer('offset', offset);
            return false;
        }
    },

    // type = 'uid' or 'offset'
    fetchBuffer: function(type, value, opts, params)
    {
        opts = opts || {};

        if (this.busyrequest) {
            this.fetchBuffer.bind(this, type, value, opts, params).defer();
            return;
        }
        this.busyrequest = true;

        // Only call onFetch() if we are loading in foreground.
        if (this.opts.onFetch && !opts.noupdate) {
            this.opts.onFetch();
        }

        params = this.addRequestParams(params);
        params.set(type, value);
        var action = this.opts.fetch_action,
            request_vals = [ type, value ];

        if (this.filter && this.filter.isFiltering()) {
            action = this.filter.getAction();
            params = this.filter.addFilterParams(params);
            // Need to capture filter params changes in the request ID
            request_vals.push(params.toJSON());
        }

        var cr = this.current_requests.get(this.id),
            id = this.id,
            req = {},
            request_id = request_vals.join('|');

        if (cr) {
            if (cr.requests.request_id) {
                // Check for repeat request.  We technically should never
                // reach here but if we do, make sure we don't go into an
                // infinite loop.
                if (++cr.requests.request_id.count == 4) {
                    this.busyrequest = false;
                    this.displayFetchError();
                    return;
                }
            } else if (type == 'offset' &&
                       cr.offset_list.indexOf(value + this.getPageSize()) != -1) {
                // Check for message list requests that are requesting
                // (essentially) the same message slice - such as two
                // scroll down requests sent in quick succession.  If the
                // original request will contain the slice needed by the
                // second request, ignore the later request.
                this.addRequest(id, cr.view_id, { offset: value, uid: null }, false);
                return;
            }
        }

        if (!opts.noupdate) {
            req.offset = (type == 'offset') ? value : null;
            req.uid = (type == 'uid') ? value : null;
        }

        if (type == 'offset') {
            // Ignore duplicates -> uniq() is an expensive call.
            req.offset_list = (cr ? cr.offset_list : []).concat($A($R(value, value + Math.min(value + params.get('buffer_size'), this.getMetaData('total_rows')))));
        }

        req.view_id = request_id;
        this.addRequest(id, request_id, req, opts.noupdate);

        this.opts.ajaxRequest(action, params, function(r) { this.ajaxUpdate(r, id, request_id); }.bind(this), Object.clone(this.opts));
        this.handleWait();
    },

    // Returns a Hash object
    addRequestParams: function(args)
    {
        var params = this.additional_params.clone(),
            cid = this.getMetaData('cacheid');
        if (cid) {
            params.update({ cacheid: cid });
        }
        return params.merge(args).merge({ buffer_size: this.buffer.bufferSize(), offset: this.currentOffset() });
    },

    ajaxUpdate: function(r, id, r_id)
    {
        if (this.busyrequest) {
            this.ajaxUpdate.bind(this, r, id, r_id).defer();
            return;
        }

        this.busyrequest = true;
        this.clearWait();

        var cr = Object.clone(this.current_requests.get(id));
        var noupdate = cr.requests[r_id].noupdate;

        if (!r.response || !r.response.rows) {
            if (id == this.id && !cr.noupdate) {
                this.displayFetchError();
            }
            this.removeRequest(id, r_id);
            return;
        }

        r.response.viewport_done = true;
        if (this.viewport_init) {
            this.viewport_init = 2;
        }

        var d = r.response.rows;
        var showing = (this.id == d.id || d.search);
        var buffer, offset;

        buffer = (showing) ? this.buffer : this.views.get(d.id).buffer;

        // Parse parameters
        d.msglist = $H(d.msglist);
        buffer.update(d.msglist, d.msgdata, d.update);
        buffer.setMetaData($H(d.other).merge({
            cacheid: d.cacheid,
            label: d.label,
            total_rows: d.totalrows
        }));

        this.removeRequest(id, r_id);

        // Don't update the view if we are now in a different view, or if
        // we are loading in the background.
        if (!showing || noupdate || cr.view_id != r_id) {
            return;
        }

        if (cr.offset !== null) {
            offset = cr.offset;
        } else {
            // If offset is null, we originally searched by ID. Thus, we
            // need to determine the offset by what was returned from the
            // server.
            offset = d.msglist.index(d.uid);
            offset = (Object.isUndefined(offset)) ? 0 : (offset - 1);
        }

        if (!this.updateContent(offset)) {
           return;
        }

        if (this.opts.onComplete) {
            this.opts.onComplete();
        }
    },

    addRequest: function(id, r_id, params, noupdate)
    {
        var req = this.current_requests.get(id);
        req = (req) ? $H(req).merge(params).toObject() : params;

        if (!req.requests) {
            req.requests = {};
        }
        req.requests[r_id] = { noupdate: noupdate, count: 1 };

        this.current_requests.set(id, req);
        this.busyrequest = false;
    },

    removeRequest: function(id, r_id)
    {
        var cr = this.current_requests.get(id);
        if (cr) {
            if ($A(cr.requests).size() == 1) {
                this.current_requests.unset(id);
            } else {
                delete cr.requests[r_id];
                this.current_requests.update(id, cr);
            }
        }
        this.busyrequest = false;
    },

    updateContent: function(offset, opts)
    {
        if (!this.buffer.isLoaded(offset)) {
            this.fetchBuffer('offset', offset, opts);
            return false;
        }

        var c = this.opts.content, c_ids = [], c_remove = [], c_rows = [];

        this.last_page_size = this.getPageSize();
        this.last_start_pos = (this.getMetaData('total_rows') > this.last_page_size) ? offset : 0;

        var newrows = [],
            rows = this.getRows(this.last_start_pos + 1, this.last_page_size);

        // Code for viewport that only needs to be initialized once.
        if (!this.firstcontent) {
            this.firstcontent = true;
            if (this.opts.onFirstContent) {
                this.opts.onFirstContent();
            }
            this.clearMsgList();
        } else {
            c_rows = c.immediateDescendants();
            c_ids = c_rows.pluck('id');
            c_rows.invoke('remove');
        }

        $R(0, rows.size() - 1).each(function(i) {
            var r;
            if (!this.force) {
                var key = c_ids.indexOf(this.opts.msgid_prefix + rows[i].uid);
                if (key != -1) {
                    r = c_rows[key];
                    c_remove.push(key);
                }
            }
            if (!r) {
                r = new Element('DIV').update(this.template.evaluate(rows[i])).firstDescendant();
                newrows.push(r);
            }
            c.appendChild(r);
        }.bind(this));

        c_remove.each(function(k) {
            c_rows[k] = null;
        });
        c_rows = c_rows.compact();

        if (c_rows.size() && this.opts.onClearRows) {
            this.opts.onClearRows(c_rows);
        }

        if (newrows.size() && this.opts.onContent) {
            this.opts.onContent(newrows);
        }

        if (this.opts.empty && !c.immediateDescendants().size()) {
            // If loading a viewport for the first time, show a blank
            // viewport rather than the empty viewport status message.
            c.update((this.viewport_init == 1) ? '' : this.opts.empty.innerHTML);
        }

        this.scroller.updateSize();
        this.scroller.moveScroll(this.last_start_pos);

        return true;
    },

    // If true, forces an update of all visible rows on viewport redraw.
    forceUpdate: function(force)
    {
        this.force = force;
    },

    displayFetchError: function()
    {
        if (this.opts.onFail) {
            this.opts.onFail();
        }
        if (this.opts.error) {
            this.opts.content.update(this.opts.error.innerHTML);
        }
    },

    checkCache: function()
    {
        this.opts.ajaxRequest(this.opts.cachecheck_action, this.addRequestParams({}), this.checkCacheCallback.bind(this));
    },

    checkCacheCallback: function(r)
    {
        // No need to do anything if ajaxUpdate() has already processed the
        // data.
        if (!r.response || !r.response.rows || r.response.viewport_done) {
            return;
        }
        if (this.isbusy) {
            this.checkCacheCallback.bind(this, r).defer();
        } else {
            var offset = this.currentOffset();
            var request_id = [ 'cache', offset ].join('|');
            this.addRequest(this.id, request_id, { offset: offset, view_id: request_id });
            this.ajaxUpdate(r, this.id, request_id);
        }
    },

    handleWait: function(call)
    {
        this.clearWait();

        // Server did not respond in defined amount of time.  Alert the
        // callback function and set the next timeout.
        if (call && this.opts.onWait) {
            this.opts.onWait();
        }

        // Call wait handler every x seconds
        if (this.opts.viewport_wait) {
            this.waitHandler = this.handleWait.bind(this, true).delay(this.opts.viewport_wait);
        }
    },

    clearWait: function()
    {
        if (this.waitHandler) {
            clearTimeout(this.waitHandler);
            this.waitHandler = null;
        }
    },

    clearMsgList: function()
    {
        var c = this.opts.content;

        if (this.opts.onClearRows) {
            this.opts.onClearRows(c.immediateDescendants());
        }

        $A(c.childNodes).each(function(n) {
            c.removeChild(n);
        });
    },

    clear: function()
    {
        this.clearMsgList();
        this.scroller.clear();
    },

    visibleRows: function()
    {
        return this.opts.content.immediateDescendants();
    },

    getRows: function(start, length)
    {
        return this.buffer.getRows(start, length);
    },

    getRowDiv: function(id, view)
    {
        if (view && view != this.id) {
            return null;
        }
        var row = this.getRowById(id);
        return (row) ? $(this.opts.msgid_prefix + row.uid) : null;
    },

    getMetaData: function(id)
    {
        return this.buffer.getMetaData(id);
    },

    setMetaData: function(vals, id)
    {
        var md;
        if (id && id != this.id) {
            if (!this.views.get(id)) {
                return;
            }
            md = this.views.get(id).buffer;
        } else {
            md = this.buffer;
        }
        md.setMetaData(vals);
    },

    currentOffset: function()
    {
        return this.scroller.currentOffset();
    },

    updateFlag: function(ids, flag, add, permanent, view)
    {
        ids.each(function(id) {
            var row = this.getRowById(id, view);
            if (!row) {
                return;
            }

            if (permanent) {
                row.bg = (add) ? $w(row.bg).concat([ flag ]).join(' ') : $w(row.bg).without([ flag ]).join(' ');
            }

            var div = this.getRowDiv(id, view);
            if (div) {
                if (add) {
                    div.addClassName(flag);
                } else {
                    div.removeClassName(flag);
                }
            }
        }.bind(this));

        if (!Object.isUndefined(view) &&
            ((this.filter && this.filter.isFiltering()) ||
             (view != this.id))) {
            this.updateFlag(ids, flag, add, permanent);
        }
    },

    getBufferSize: function()
    {
        return this.buffer.size;
    },

    getLineHeight: function()
    {
        if (this.line_height) {
            return this.line_height;
        }

        // To avoid hardcoding the line height, create a temporary row to
        // figure out what the CSS says.
        var d = new Element('DIV', { className: this.opts.content_class }).insert(new Element('DIV', { className: this.opts.row_class })).hide();
        document.body.appendChild(d);
        this.line_height = d.getHeight();
        document.body.removeChild(d);

        return this.line_height;
    },

    setPageSize: function(size)
    {
        this.page_size = size;
        this.renderViewport();
    },

    getPageSize: function()
    {
        return this.page_size;
    },

    getCurrPageSize: function()
    {
        return Math.min(this.getPageSize(), this.getMetaData('total_rows'));
    },

    getDefaultPageSize: function()
    {
        return Math.max(parseInt(this.getMaxPageSize() * 0.45), 5);
    },

    getMaxPageSize: function()
    {
        return parseInt(this.opts.maxHeight() / this.getLineHeight());
    },

    showSplitPane: function(show)
    {
        this.show_split_pane = show;
        this.onResize();
    },

    renderViewport: function(noupdate)
    {
        if (!this.viewport_init) {
            return;
        }

        var panesize;

        // Get split pane dimensions
        if (this.opts.split_pane) {
            var lh = this.getLineHeight(),
                pane = $(this.opts.split_pane);
            if (this.show_split_pane) {
                if (!pane.visible()) {
                    this.initSplitBar();
                    this.page_size = (this.splitbar_loc) ? this.splitbar_loc : this.getDefaultPageSize();
                }
                panesize = this.opts.maxHeight() - (lh * (this.page_size + 1));
                $(this.opts.content).setStyle({ height: (lh * this.page_size) + 'px' });
            } else {
                if (pane.visible()) {
                    this.splitbar_loc = this.page_size;
                    $(pane, this.splitbar).invoke('hide');
                }
                this.page_size = this.getMaxPageSize();
            }
        }

        $(this.opts.content).setStyle({ height: (lh * this.page_size) + 'px' });
        if (panesize) {
            pane.setStyle({ height: panesize + 'px' });
            $(pane, this.splitbar).invoke('show');
        }

        if (!noupdate) {
            this.scroller.onResize();
        }
    },

    initSplitBar: function()
    {
        if (this.splitbar) {
            return;
        }

        this.splitbar = $(this.opts.splitbar);
        new Draggable(this.opts.splitbar, {
            constraint: 'vertical',
            ghosting: true,
            onStart: function() {
                // Cache these values since we will be using them multiple
                // times in snap().
                var lh = this.getLineHeight();
                this.sp = { lh: lh, pos: $(this.opts.content).positionedOffset()[1], max: parseInt((this.opts.maxHeight() - 100) / lh) };
            }.bind(this),
            snap: function(x, y) {
                var l = parseInt((y - this.sp.pos) / this.sp.lh);
                if (l < 1) {
                    l = 1;
                } else if (l > this.sp.max) {
                    l = this.sp.max;
                }
                this.sp.lines = l;
                return [ 0, this.sp.pos + (l * this.sp.lh) ];
            }.bind(this),
            onEnd: function() {
                $(this.opts.splitbar).setStyle({ top: 0 });
                this.setPageSize(this.sp.lines);
            }.bind(this)
        });
    }

});

/**
 * ViewPort_Scroller
 */
var ViewPort_Scroller = Class.create({

    initialize: function(parent)
    {
        this.parent = parent;

        // Initialize other variables
        this.pagesize = this.scrollsize = this.totalrows = null;
    },

    createScrollBar: function()
    {
        if (this.scrollDiv) {
            return;
        }
        var c = this.parent.opts.content;

        // Create the outer div.
        this.scrollsize = c.getHeight();
        this.scrollDiv = new Element('DIV', { className: 'sbdiv', style: 'height:' + this.scrollsize + 'px;' }).hide();

        // Create the cursor element.
        this.cursorDiv = new Element('DIV', { className: 'sbcursor' });

        // Create scrollbar object.
        this.scrollbar = new DimpSlider(this.cursorDiv, this.scrollDiv, { axis: 'vertical', onClickTrack: this.onClickTrackHandler.bind(this), onChange: this.onScroll.bind(this), onSlide: this.parent.opts.onSlide ? this.parent.opts.onSlide : null });

        // Add scrollbar to parent viewport
        this.scrollDiv.appendChild(this.cursorDiv);
        c.parentNode.insertBefore(this.scrollDiv, c.nextSibling);

        // Give our parent a right margin just big enough to
        // accomodate the scrollbar.
        c.setStyle({ marginRight: '-' + this.scrollDiv.getWidth() + 'px' });

        // Mouse wheel handler.
        c.observe(Prototype.Browser.IE ? 'mousewheel' : 'DOMMouseScroll', function(e) {
            this.moveScroll(this.currentOffset() + ((e.wheelDelta >= 0 || e.detail < 0) ? -3 : 3));
            if (this.parent.opts.onMouseScroll) {
                this.parent.opts.onMouseScroll(e);
            }
            e.stop();
        }.bindAsEventListener(this));
    },

    onResize: function()
    {
        if (!this.scrollDiv) {
            return;
        }

        // Update the container div.
        this.scrollsize = this.parent.opts.content.getHeight();
        this.scrollDiv.setStyle({ height: this.scrollsize + 'px' });

        // Update the scrollbar size.
        this.scrollbar.updateTrackLength();

        // Update the cursor size
        this.updateSize();

        // Update displayed content.
        this.parent.refresh();
        this.onScroll();
    },

    updateSize: function()
    {
        this.createScrollBar();
        this.pagesize = this.parent.getPageSize();
        this.totalrows = this.parent.getMetaData('total_rows');
        if (this.totalrows <= this.pagesize) {
            this.parent.opts.content.setStyle({ cssFloat: 'none' });
            this.scrollDiv.hide();
        } else {
            this.scrollDiv.show();
            this.parent.opts.content.setStyle({ cssFloat: 'left' });
            // Minimum cursor size = 10px
            var l = Math.max(10, Math.round((this.pagesize / this.totalrows) * this.scrollsize));
            this.cursorDiv.setStyle({ height: l + 'px' });
            this.scrollbar.setHandleLength(l);
        }
    },

    clear: function()
    {
        if (this.scrollDiv) {
            this.totalrows = 0;
            this.scrollDiv.hide();
        }
    },

    moveScroll: function(offset)
    {
        if (this.scrollDiv && this.scrollDiv.visible()) {
            this.scrollbar.setScrollPosition(this.offsetVal(offset));
            this.onScroll();
        }
    },

    offsetVal: function(offset)
    {
        var offset_rows = this.totalrows - this.pagesize;
        return Math.min(Math.max(offset, 0), offset_rows) / offset_rows;
    },

    onClickTrackHandler: function(v)
    {
        var dir = ((v - this.scrollbar.translateToPx(this.scrollbar.getValue()).replace(/px$/,"")) < 0) ? -1 : 1;
        this.moveScroll(this.currentOffset() - (1 * dir) + (this.pagesize * dir));
        this.scrollbar.clearClick();
    },

    onScroll: function()
    {
        if (this.parent.opts.onScroll) {
            this.parent.opts.onScroll();
        }

        this.parent.requestContentRefresh(this.currentOffset());

        if (this.parent.opts.onScrollIdle) {
            this.parent.opts.onScrollIdle();
        }
    },

    currentOffset: function()
    {
        return (!this.scrollDiv || !this.scrollDiv.visible()) ? 0 : Math.max(Math.round(this.scrollbar.getValue() * (this.totalrows - this.pagesize)), 0);
    }

});

/**
 * ViewPort_Buffer
 *
 * Note: recognize the difference between offset (current location in the
 * viewport - starts at 0) with start parameters (the row numbers - starts
 * at 1).
 *
 * Also, due to limitations in PHP casting an array to an object, the keys
 * of this.list will be strings.  Always convert to integer before using.
 * We don't do this in PHP since casting from an array to an object is much
 * less memory/resource intensive than looping through the array.
 */
var ViewPort_Buffer = Class.create({

    initialize: function(parent, b_pages, l_factor)
    {
        this.clear();
        this.bufferPages = b_pages;
        this.data = $H({ total_rows: 0 });
        this.limitFactor = l_factor;
        this.parent = parent;
    },

    limitTolerance: function()
    {
        return Math.round(this.bufferSize() * (this.limitFactor / 100));
    },

    bufferSize: function()
    {
        return Math.round(this.bufferPages * this.parent.getMaxPageSize());
    },

    update: function(list, data, update)
    {
        this.data.update(data);
        this.list = (update) ? $H(list) : this.list.merge(list);
        this.size = this.list.size();
        this.lookup = this.list.flip();
    },

    isLoaded: function(offset)
    {
        return $R(offset + 1, Math.min(offset + this.parent.getMaxPageSize() - 1, this.getMetaData('total_rows'))).all(function(o) {
            return !Object.isUndefined(this.list.get(o));
        }.bind(this));
    },

    isNearingTopLimit: function(offset)
    {
        if (offset == 0) {
            return false;
        }
        return $R(Math.max(offset + 1 - this.limitTolerance(), 1), offset).any(function(o) {
            return Object.isUndefined(this.list.get(o));
        }.bind(this));
    },

    isNearingBottomLimit: function(offset)
    {
        // Search for missing messages in reverse order since in normal usage
        // (sequential scrolling through the message list) messages are
        // more likely to be missing at furthest from the current view.
        return $A($R(offset + 1, Math.min(offset + this.limitTolerance() + this.parent.getMaxPageSize() - 1, this.getMetaData('total_rows')))).reverse().any(function(o) {
            return Object.isUndefined(this.list.get(o));
        }.bind(this));
    },

    isNearingLimit: function(offset)
    {
        if (this.size != this.getMetaData('total_rows')) {
            if (this.isNearingTopLimit(offset)) {
                return 'top';
            } else if (this.isNearingBottomLimit(offset)) {
                return 'bottom';
            }
        }
        return false;
    },

    getRowById: function(id)
    {
        var o = this.inBuffer(id);
        return (o) ? this.getRows(parseInt(o), 1).first() : null;
    },

    getRows: function(start, length)
    {
        return $R(start, Math.min(start + length - 1, this.getMetaData('total_rows'))).collect(function(num) {
            if (!Object.isUndefined(this.list.get(num))) {
                var uid = this.list.get(num);
                // We can directly write the rownum to the original object
                // since we will always rewrite when creating rows.
                this.data.get(uid).rownum = num;
                return this.data.get(uid);
            }
        }.bind(this)).compact();
    },

    inBuffer: function(id)
    {
        return this.lookup.get(id);
    },

    removeByUid: function(uids)
    {
        var newlist = $H(), rowsubtract = 0, uidremove = [];
        this.list.keys().numericSort().each(function(n) {
            var uid = this.list.get(n);
            var key = uids.indexOf(uid);
            if (key == -1) {
                var newrow = n - rowsubtract;
                newlist.set(newrow, uid);
                this.lookup.set(uid, newrow);
            } else {
                uidremove.push(uid);
                ++rowsubtract;
            }
        }.bind(this));
        [ this.data, this.lookup ].invoke('unset', uidremove);
        this.list = newlist;
        this.size = this.list.size();
    },

    clear: function()
    {
        this.data = $H();
        this.list = $H();
        this.lookup = $H();
        this.size = 0;
    },

    getMetaData: function(id)
    {
        return this.data.get(id);
    },

    setMetaData: function(vals)
    {
        this.data.update(vals);
    }

});

/**
 * ViewPort_Filter
 */
var ViewPort_Filter = Class.create({

    initialize: function(parent, action, callback)
    {
        this.parent = parent;
        this.action = action;
        this.callback = callback;

        // Initialize other variables
        this.filterid = 0;
        this.filtering = this.last_filter = this.last_folder = this.last_folder_params = null;
    },

    filter: function(val, params)
    {
        params = params || {};
        val = (val === null) ? this.last_filter : val;
        if (!val) {
            this.clear();
            return;
        }

        val = this.last_filter = val.toLowerCase();

        if (this.filtering) {
            this.parent.fetchBuffer('offset', 0, {}, params);
            return;
        }

        this.filtering = ++this.filterid + '%search%';
        this.last_folder = this.parent.id;
        this.last_folder_params = this.parent.additional_params.merge(params);

        // Filter visible rows immediately.
        var c = this.parent.opts.content;
        var delrows = c.immediateDescendants().findAll(function(n) {
            return n.getText(true).toLowerCase().indexOf(val) == -1;
        });
        if (this.parent.opts.onClearRows) {
            this.parent.opts.onClearRows(delrows);
        }
        delrows.invoke('remove');
        this.parent.scroller.clear();
        if (this.parent.opts.empty && !c.immediateDescendants().size()) {
            c.update(this.parent.opts.empty.innerHTML);
        }

        this.parent.loadView(this.filtering, this.last_folder_params);
    },

    isFiltering: function()
    {
        return this.filtering;
    },

    getAction: function()
    {
        return this.action;
    },

    // params is a Hash object
    addFilterParams: function(params)
    {
        if (!this.filtering) {
            return params;
        }

        params.update({ filter: this.last_filter });

        // Get parameters from a callback function, if defined.
        if (this.callback) {
            params.update(this.callback());
        }

        return params;
    },

    clear: function(reset)
    {
        if (this.filtering) {
            if (!reset) {
                this.parent.loadView(this.last_folder, this.last_folder_params);
            }
            this.parent.deleteView(this.filtering);
            this.filtering = this.last_filter = this.last_folder = null;
        }
    }

});
