<?php
/**
 * ShowMessage view logic.
 *
 * $Horde: dimp/lib/Views/ShowMessage.php,v 1.70.2.2 2008/01/03 07:07:56 slusarz Exp $
 *
 * Copyright 2005-2008 The Horde Project (http://www.horde.org/)
 *
 * See the enclosed file COPYING for license information (GPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/gpl.html.
 *
 * @package DIMP
 */

require_once IMP_BASE . '/lib/IMAP/MailboxCache.php';
require_once IMP_BASE . '/lib/MIME/Contents.php';
require_once IMP_BASE . '/lib/MIME/Headers.php';
require_once IMP_BASE . '/lib/UI/Message.php';

class DIMP_Views_ShowMessage {

    /**
     * @var string
     */
    var $_imagedir;

    /**
     * Constructor.
     */
    function DIMP_Views_ShowMessage()
    {
        $this->_imagedir = $GLOBALS['registry']->getImageDir('dimp');
    }

    /**
     * Builds a string containing a list of addresses.
     *
     * @access private
     *
     * @param IMP_Headers &$headers  The headers object.
     * @param string $field          The address field to parse.
     * @param boolean $set           Set the associated header with the return
     *                               string?
     * @param boolean $popup         Are we viewing in a popup?
     *
     * @return string  String containing the formatted address list.
     */
    function _buildAddressLinks(&$headers, $field, $set = false, $popup = false)
    {
        /* Make sure this is a valid object address field. */
        $array = $headers->getOb($field);
        if (empty($array) || !is_array($array)) {
            return;
        }

        $addr_array = array();

        foreach ($headers->getAddressesFromObject($array) as $ob) {
            if (empty($ob->address) || empty($ob->inner)) {
                continue;
            }

            /* If this is an incomplete e-mail address, don't link to
             * anything. */
            if (!empty($GLOBALS['conf']['hooks']['addressformatting'])) {
                $addr_array[] = Horde::callHook('_dimp_hook_addressformatting', array($ob), 'dimp');
            } elseif (stristr($ob->host, 'UNKNOWN') !== false) {
                $addr_array[] = $ob->address;
            } else {
                $addr_array[] = '<a class="address" href="#" personal="' . htmlspecialchars($ob->personal) . '" email="' . htmlspecialchars($ob->inner) . '" address="' . htmlspecialchars($ob->address) . '">' . htmlspecialchars($ob->address) . Horde::img('popdown.png', '', array(), $this->_imagedir) . '</a>';
            }
        }

        /* If left with an empty address list, inform the user that the
         * recipient list is purposely "undisclosed". */
        if (empty($addr_array)) {
            $ret = _("Undisclosed Recipients");
        } else {
            /* Build the address line. */
            if (count($addr_array) > 20) {
                $ret = '<div id="at_' . $field . '">' .
                       Horde::link('#', '', 'widget', '', 'toggleAddressesBlock(\'' . $field . '\', \'' . count($addr_array) . '\'); return false;', '', '') .
                       sprintf(_("[Show addresses - %s recipients]"), count($addr_array)) . '</a></div>' .
                       '<div id="ab_' . $field . '" style="display:none;">' . implode(', ', $addr_array) . '</div>';
            } else {
                $ret = implode(', ', $addr_array);
            }
        }

        /* Set the header value, if requested. */
        if ($set) {
            $headers->setValue($field, $ret);
        }

        return $ret;
    }

    /**
     * $args: folder, index, fullmessage, options
     */
    function showMessage($args)
    {
        $fullmessage = !empty($args['fullmessage']);
        $folder = $args['folder'];
        $index = $args['index'];

        /* Get the IMP_Headers:: object. */
        $mailbox_cache = &IMP_MailboxCache::singleton();
        $cache_entry = $mailbox_cache->retrieve($folder, array($index), 1 | 32);
        $ob = reset($cache_entry);
        if ($ob === false) {
            $GLOBALS['notification']->push(_("Requested message not found."), 'horde.error');
            return false;
        }

        $imp_headers = &Util::cloneObject($ob->header);

        /* Parse MIME info and create the body of the message. */
        $imp_contents = &IMP_Contents::singleton($index . IMP_IDX_SEP . $folder);
        if (is_a($imp_contents, 'PEAR_Error') ||
            !$imp_contents->buildMessage()) {
            $GLOBALS['notification']->push(_("Requested message not found."), 'horde.error');
            return false;
        }

        /* Get the IMP_UI_Message:: object. */
        $imp_ui = new IMP_UI_Message();

        /* Update the message flag, if necessary. */
        if (($_SESSION['imp']['base_protocol'] == 'imap') && empty($ob->seen)) {
            require_once IMP_BASE . '/lib/Mailbox.php';
            require_once IMP_BASE . '/lib/Message.php';
            $imp_mailbox = &IMP_Mailbox::singleton($folder, $index);
            $imp_message = &IMP_Message::singleton();
            $imp_message->flag(array('seen'), $imp_mailbox, true);
        }

        /* Determine if we should generate the attachment strip links or
         * not. */
        if ($GLOBALS['prefs']->getValue('strip_attachments')) {
            $imp_contents->setStripLink(true);
        }

        /* Show summary links. */
        $imp_contents->showSummaryLinks(true);

        $attachments = $imp_contents->getAttachments();
        $msgText = $imp_contents->getMessage();

        /* Build From address links. */
        if (!$fullmessage) {
            $temp_headers = Util::cloneObject($imp_headers);
        }
        $this->_buildAddressLinks($imp_headers, 'from', true, $fullmessage);

        /* Build To/Cc/Bcc links. */
        foreach (array('to', 'cc', 'bcc') as $val) {
            $msgAddresses[] = $imp_headers->getValue($val);
            $this->_buildAddressLinks($imp_headers, $val, true, $fullmessage);
        }

        /* Build Reply-To address links. */
        if (($reply_to = $this->_buildAddressLinks($imp_headers, 'reply-to', false, $fullmessage))) {
            if (!($from = $imp_headers->getValue('from')) || ($from != $reply_to)) {
                $imp_headers->setValue('Reply-to', $reply_to);
            } else {
                $imp_headers->removeHeader('reply-to');
            }
        }

        /* Develop the list of Headers to display now. We will deal with the
         * 'basic' header information first since there are various
         * manipulations we do to them. */
        $headers_list = $imp_ui->basicHeaders();
        if (empty($args['headers'])) {
            $args['headers'] = array('from', 'date', 'to', 'cc');
        }
        // TODO: When PHP 5.1.0+ is required, we can use array_intersect_key.
        $basic_headers = array();
        foreach (array_intersect(array_keys($headers_list), $args['headers']) as $v) {
            $basic_headers[$v] = $headers_list[$v];
        }

        $headers = array();
        $header_from = '';
        foreach ($basic_headers as $head => $str) {
            if ($val = $imp_headers->getValue($head)) {
                if ($head == 'date') {
                    /* Add local time to date header. */
                    $val = nl2br($imp_headers->addLocalTime(htmlspecialchars($val)));
                } elseif ($head == 'from') {
                    $header_from = $val;
                } elseif (!in_array($head, array('to', 'cc', 'bcc', 'reply-to'))) {
                    $val = htmlspecialchars($val);
                }
                $headers[] = array('id' => String::ucfirst($head), 'name' => $str, 'value' => $val);
            }
        }

        /* Display the user-specified headers for the current identity. */
        $user_hdrs = $imp_ui->getUserHeaders();
        if (!empty($user_hdrs)) {
            $full_h = $imp_headers->getAllHeaders();
            foreach ($user_hdrs as $user_hdr) {
                foreach ($full_h as $head => $val) {
                    if (stristr($head, $user_hdr) !== false) {
                        $headers[] = array('name' => $head, 'value' => htmlspecialchars($val));
                    }
                }
            }
        }

        /* Process the subject. */
        $subject_name = _("Subject");
        if (($subject = $imp_headers->getValue('subject'))) {
            require_once 'Horde/Text.php';
            $subject = $subject_orig = Text::htmlSpaces(IMP::filterText($subject));
        } else {
            $subject = $subject_orig = htmlspecialchars(_("[No Subject]"));
        }
        $subject_escaped = rawurlencode(@html_entity_decode($subject, ENT_QUOTES, NLS::getCharset()));

        /* See if the 'X-Priority' header has been set. */
        switch ($imp_headers->getXpriority()) {
        case 'high':
            $subject = Horde::img('mail_priority_high.png', _("High Priority")) . ' ' . $subject;
            break;

        case 'low':
            $subject = Horde::img('mail_priority_low.png', _("Low Priority")) . ' ' . $subject;
            break;
        }

        $atc_display = $GLOBALS['prefs']->getValue('attachment_display');
        $show_parts = (!empty($attachments) && (($atc_display == 'list') || ($atc_display == 'both')));
        $downloadall_link = $imp_contents->getDownloadAllLink();

        /* Add attachment info. */
        if ($attachments && ($show_parts || $downloadall_link)) {
            $val = $imp_contents->getAttachmentSummary();
            if ($downloadall_link) {
                $val .= ' ' . Horde::link($downloadall_link) . _("Save All") . '</a>';
            }
            if ($show_parts) {
                $val .= '<table id="partlist" style="display:none" cellspacing="2">' . $attachments . '</table>';
                $headers[] = array('name' => Horde::link('#', '', '', '', '$(\'partlist\', \'partlist_col\', \'partlist_exp\').invoke(\'toggle\');return false;') .
                                             Horde::img('arrow_collapsed.png', '', array('id' => 'partlist_col'), $this->_imagedir) .
                                             Horde::img('arrow_expanded.png', '', array('id' => 'partlist_exp', 'style' => 'display:none'), $this->_imagedir) .
                                             ' ' . Horde::img('attachment.png') . '</a>',
                                   'value' => $val);
                    } else {
                $headers[] = array('name' => Horde::img('attachment.png'), 'value' => $val);
            }
        }

        /* Set DIMP options. */
        $newwin_img = $newwin_js = '';
        $options = array();

        // Open in new window
        if (in_array('new_window', $args['options'])) {
            $newwin_img = Horde::img('newwin.png', _("Open in new window"), '', $this->_imagedir);
            $newwin_js = 'return DimpCore.msgWindow({folder: \'' . htmlspecialchars($folder) . '\', uid: \'' . (int)$index . '\'} );';
            $options[] = array(
                'img' => $newwin_img,
                'label' => _("Open in new window"),
                'onclick' => $newwin_js
            );
        }

        // Print in new window
        if (in_array('print', $args['options'])) {
            $options[] = array(
                'img' => Horde::img('print.png', '', '', $GLOBALS['registry']->getImageDir('horde')),
                'label' => _("Print"),
                'onclick' => 'window.print(); return false;'
            );
        }

        // Message Source link
        if (in_array('msg_source', $args['options'])) {
            $base_part = $imp_contents->getMIMEMessage();
            $ms_url = Util::addParameter($imp_contents->urlView($base_part, 'view_source'), 'thismailbox', $folder);
            if (!($ms_id = $base_part->getMIMEId())) {
                $ms_id = abs(crc32(serialize($base_part)));
            }
            $options[] = array(
                'img' => Horde::img('message_source.png', _("View Message Source"), '', $this->_imagedir),
                'label' => _("View Message Source"),
                'onclick' => 'view(\'' . $ms_url . '\', \'' . $ms_id . '\'); return false;'
            );
        }

        /* Store all template information in an array so it can be possibly be
         * altered via the messageview hook. */
        $t_array = array(
            'subject' => $subject,
            'subject_name' => $subject_name,
            'headers' => $headers,
            'message' => $msgText,
            'options' => $options,
            'compose'=> array('img' => Horde::img('compose.png', '', '', $this->_imagedir),
                              'label' => _("New Message")),
            'addcontact' => array('img' => Horde::img('add_contact.png', '', '', $this->_imagedir),
                                  'label' => _("Add to Address Book")),
            'notfullmessage' => !$fullmessage
        );

        if (!$fullmessage) {
            $curr_time = time();
            $curr_time -= $curr_time % 60;
            $ltime_val = localtime();
            $today_start = mktime(0, 0, 0, $ltime_val[4] + 1, $ltime_val[3], 1900 + $ltime_val[5]);
            $today_end = $today_start + 86400;
            if (empty($ob->date)) {
                $udate = false;
            } else {
                $ob->date = preg_replace('/\s+\(\w+\)$/', '', $ob->date);
                $udate = strtotime($ob->date, $curr_time);
            }
            if ($udate === false || $udate === -1) {
                $header_date = _("Unknown Date");
            } elseif (($udate < $today_start) || ($udate > $today_end)) {
                /* Not today, use the date. */
                $header_date = strftime($GLOBALS['prefs']->getValue('date_format'), $udate);
            } else {
                /* Else, it's today, use the time. */
                $header_date = strftime($GLOBALS['prefs']->getValue('time_format'), $udate);
            }

            $t_array = array_merge($t_array, array(
                'arrow_expanded' => $this->_imagedir . '/arrow_expanded.png',
                'arrow_collapsed' => $this->_imagedir . '/arrow_collapsed.png',
                'newwin_img' => $newwin_img,
                'newwin_js' => $newwin_js,
                'header_attach' => ($attachments && ($show_parts || $downloadall_link)) ?  Horde::img('attachment.png') : '',
                'header_date' => $header_date,
                'header_from' => $header_from,
            ));
        }

        if (!empty($GLOBALS['conf']['hooks']['messageview'])) {
            $t_array = Horde::callHook('_dimp_hook_messageview', array($t_array, $fullmessage), 'dimp');
        }

        $result = new stdClass;
        $result->folder = $folder;
        $result->index = $index;
        $result->subject = $subject_orig;

        require_once IMP_BASE . '/lib/Template.php';
        $t = new IMP_Template(DIMP_TEMPLATES . '/imp/');
        $t->setOption('gettext', true);
        foreach ($t_array as $key => $val) {
            $t->set($key, $val);
        }
        $result->msg = $t->fetch('message.html');

        // Add javascript code
        $js_code = array();
        if ($fullmessage) {
            $js_code[] = 'DimpCore.setTitle(decodeURIComponent(\'' . $subject_escaped . '\'))';
        }

        /* Retrieve any history information for this message. */
        if (!empty($GLOBALS['conf']['maillog']['use_maillog'])) {
            if ($fullmessage) {
                require_once IMP_BASE . '/lib/Maillog.php';
                IMP_Maillog::displayLog($imp_headers->getValue('message-id'));
            }

            /* Do MDN processing now. */
            if ($imp_ui->MDNCheck($ob->header)) {
                $confirm_link = Horde::link('#', '', '', '', 'DimpCore.doAction(\'SendMDN\',{folder:\'' . $folder . '\',index:' . $index . '}); return false;', '', '') . _("HERE") . '</a>';
                $GLOBALS['notification']->push(sprintf(_("The sender of this message is requesting a Message Disposition Notification from you when you have read this message. Click %s to send the notification message."), $confirm_link), 'dimp.request', array('content.raw'));
            }

            $notify = DIMP::notify();
            if (!empty($notify)) {
                $js_code[] = $notify;
            }
        }

        IMP::addInlineScript($js_code, true);

        return $result;
    }

}
